#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#define PREFIX "/ade/share/a2ixlibrary"

struct liblist {
  struct liblist *next;
  char *name;
  char *cname;
};

struct liblist *head = NULL;

static void safename(char *p)
{
  if (!isalpha(*p))
    *p = '_';
  p++;
  while (*p)
  {
    if (!isalnum(*p))
      *p = '_';
    p++;
  }
}

static void add_lib(char *name)
{
  struct liblist *lib;

  for (lib = head; lib; lib = lib->next)
    if (!stricmp(lib->name, name))
      return;

  lib = malloc(sizeof(struct liblist));
  
  if (!lib)
  {
    fprintf(stderr, "Out of memory\n");
    exit(1);
  }
  lib->name = name;
  lib->cname = strdup(name);
  safename(lib->cname);
  lib->next = head;
  head = lib;
}

void cat(char *fname, FILE *x)
{
#define BUFSIZE 65536
  FILE *in;
  static char buf[BUFSIZE];
  int bytes;
  
  in = fopen(fname, "r");
  if (in == NULL)
  {
    fprintf(stderr, "cannot open %s for reading\n", fname);
    exit(1);
  }
  while (!feof(in) && (bytes = fread(buf, 1, BUFSIZE, in)))
    fwrite(buf, 1, bytes, x);
  fclose(in);
}

int main(int argc, char **argv)
{
  FILE *out, *x;
  struct liblist *lib;

  while (--argc)
  {
    char *arg = *++argv;
    int len = strlen(arg);
    
    if (len > 2 && !memcmp(arg, "-l", 2))
    {
      add_lib(arg + 2);
    }
    else if (len > 2 && !stricmp(arg + len - 2, ".a"))
    {
      char *lib;
      
      arg[len - 2] = '\0';
      lib = strrchr(arg, '/');
      if (lib == NULL)
        lib = strrchr(arg, ':');
      if (lib == NULL)
        lib = arg - 1;
      if (!strnicmp(lib + 1, "lib", 3))
        add_lib(lib + 4);
    }
  }

  out = fopen("xlibs.c", "w");
  if (out == NULL)
  {
    fprintf(stderr, "cannot open xlibs.c for writing\n");
    exit(1);
  }
  x = fopen("shared.x", "w");
  if (x == NULL)
  {
    fprintf(stderr, "cannot open shared.x for writing\n");
    exit(1);
  }
  cat(PREFIX "/amiga_exe_script.x", x);
  for (lib = head; lib; lib = lib->next)
  {
    static char buf[256];
    
    sprintf(buf, PREFIX "/ldscripts/%s.x", lib->name);
    if (!access(buf, R_OK))
    {
      fprintf(out, "extern long %sBase; long *__p%sBase = &%sBase;\n",
	      lib->cname, lib->cname, lib->cname);
      cat(buf, x);
    }
  }
  fprintf(x, "}}\n");
  fclose(out);
  fclose(x);
  system("gcc -c xlibs.c");
  return 0;
}
